
import java.util.ArrayList;

import org.processing.wiki.triangulate.Triangle;
import org.processing.wiki.triangulate.Triangulate;

import controlP5.*;

import biga.shapes2D.Rectangle;
import biga.utils.PolygonUtils;
import biga.utils.Simplify;
import biga.utils.TriangleUtils;

import blobDetection.Blob;
import blobDetection.BlobDetection;
import blobDetection.EdgeVertex;

import processing.core.PApplet;
import processing.core.PImage;
import processing.core.PVector;


public class Detector extends PApplet
{
	
	private BlobDetection theBlobDetection;
	private PImage img;
	private PImage texture;
	
	private ControlP5 controlP5;
	
	private float scale = 2.5f;
	private float tolerance = 3f;
	private float distance = 10f; 
	
	private ArrayList<Patch> patches;
	private Slider sliderFriction;
	private Slider sliderDamping;
	private CheckBox checkboxes;
	
	private int moveMode = Patch.MOVE_TARGET;
	private int motionType = Patch.MOTION_NONE;
	
	private int patchId = 0;
	private Slider sliderRepel;
	private Slider sliderAttract;
	private Slider randomOffset;
	private Slider noiseOffset;
	private Slider sliderScale;
	private Slider sliderTolerance;
	private Slider sliderDistance;
	
	
	public void setup()
	{
		size( 640,640 );
		background( 255 );
		initApplet();
	}
	
	public void initApplet()
	{
		

		controlP5 = new ControlP5(this);
		int oy = -10;
		
		//sliderScale = controlP5.addSlider( 	"[ W ] scale", 0,5, scale, 10, oy+=20, 100,10);
		sliderTolerance = controlP5.addSlider( 	"[ W ] simplification tolerance", 	0,20,tolerance, 10, oy+=20, 100,10);
		sliderDistance = controlP5.addSlider( 	"[ W ] simplification distance", 	0,50,distance, 	10, oy+=20, 100,10);
		
		Button btn = controlP5.addButton( "re-compute", 255, 10, oy+20, 100, 20 );
		btn.setId(0);
		
		
		sliderDamping = controlP5.addSlider( 	"[ H ] damping", 			0,	1,		Patch.damping, 	10, oy+=50, 100,10);
		sliderFriction = controlP5.addSlider( 	"[ G & H ] friction", 		0,	1,		Patch.friction, 10, oy+=20, 100,10);
		
		noiseOffset = controlP5.addSlider( 		"[ F ] noise Offset", 		0,	50, 	5, 		10, oy+=20, 100,10);
		randomOffset = controlP5.addSlider( 	"[ C ] random Offset",		0,	200,	50, 	10, oy+=20, 100,10);
		sliderRepel = controlP5.addSlider( 		"[ B ] repel Distance", 	0,	200,	100, 	10, oy+=20, 100,10);
		sliderAttract = controlP5.addSlider( 	"[ N ] attract Distance", 	0,	200,	100, 	10, oy+=20, 100,10);
		
		checkboxes = controlP5.addCheckBox( "abc", 10, oy+=20 );
		checkboxes.addItem( "smooth", 0 );
		checkboxes.addItem( "solid", 0 );
		checkboxes.addItem( "wireframe", 0 );
		checkboxes.addItem( "incircles", 0 );
		checkboxes.addItem( "voronoi", 0 );
		checkboxes.addItem( "outline", 0 );
		
		checkboxes.getItem( 1 ).toggle();
		
		drawBlobsAndEdges();
		
	}
	
	public void draw()
	{
		if( !mousePressed )
		{
			
			fill( 0, 0, 0, 64 );
			rect( 0,0, width, height );
			
		}
		
		for ( Patch p: patches )
		{
			switch( motionType )
			{
				case Patch.MOTION_RANDOM: 	p.randomMove( noiseOffset.value() );		break;
				case Patch.MOTION_FOLLOW: 	p.randomize( new PVector( mouseX,mouseY ), randomOffset.value() );		break;
				case Patch.MOTION_ATTRACT: 	p.attract( new PVector( mouseX,mouseY ), ( double )sliderAttract.value() );break;
				case Patch.MOTION_REPEL: 	p.repel( new PVector( mouseX,mouseY ), ( double )sliderRepel.value() );break;
			}
			
			switch( moveMode )
			{
				case Patch.MOVE_TARGET: 	p.target();		break;
				case Patch.MOVE_ELASTIC: 	p.elastic();	break;
			}
			
			
			noFill();
			Boolean solid = ( checkboxes.getItem( 1 ).value() == 1.0 );
			if( checkboxes.getItem( 3 ).value() == 1.0 )
			{
				p.renderIncircles( this, solid );
			}
			else if ( checkboxes.getItem( 4 ).value() == 1.0)
			{
				p.renderVoronoi( this, solid );
			}
			else if( checkboxes.getItem( 1 ).value() == 1.0 )
			{
				p.render( this );
			}
			else if( checkboxes.getItem( 2 ).value() == 1.0 )
			{
				p.renderWireFrame( this );
			}
			noFill();
			if( checkboxes.getItem( 5 ).value() == 1.0 )
			{
				p.renderOutline( this, 0xFFFFFF, 2 );
			}
		}
		
	}

	public void keyPressed()
	{
		switch( key )
		{
		
			case 'x':
				motionType = Patch.MOTION_NONE; 
				for( Patch p : patches ) p.reset();		
				break;
				
			case 'c':	
				if( motionType == Patch.MOTION_FOLLOW ) motionType = Patch.MOTION_NONE;		
				else  motionType = Patch.MOTION_FOLLOW;
				break;
				
			case 'v':
				motionType = Patch.MOTION_NONE; 
				Boolean ok = true;
				while( ok )
				{
					Patch lePatch = ( Patch )( patches.get( patchId ) );
					if( lePatch.origins.size() >= 1 )
					{
						ok = false;
						break;
					}
					else
					{
						System.out.println( "->" + patchId + "   " + lePatch.origins.size() );
						patchId++;
						patchId %= patches.size();
					}
				}
				for( Patch p : patches )p.targetPatch( ( Patch )( patches.get( patchId ) ) );

				patchId++;
				patchId %= patches.size();
				break;
			
			case 'b':		
				if( motionType == Patch.MOTION_REPEL ) motionType = Patch.MOTION_NONE;		
				else  motionType = Patch.MOTION_REPEL;
				break;
				
			case 'n':	
				if( motionType == Patch.MOTION_ATTRACT ) motionType = Patch.MOTION_NONE;		
				else  motionType = Patch.MOTION_ATTRACT;	
				break;
				
			case 'd':	
				motionType = Patch.MOTION_NONE;	
				for( Patch p : patches )p.reset();
				break;
				
			case 'f':	
				if( motionType == Patch.MOTION_RANDOM ) motionType = Patch.MOTION_NONE;		
				else  motionType = Patch.MOTION_RANDOM;		
				break;
				
			case 'g':	
				moveMode = Patch.MOVE_TARGET;			
				break;
				
			case 'h':	
				for( Patch p : patches )p.resetElasticTargets();
				moveMode = Patch.MOVE_ELASTIC; 					
				break;
			
			case 'w':	
				drawBlobsAndEdges();
				break;
				
			case 's':	
				saveFrame( "frames/pp2011-####.tif" );
				break;
		}
	}
	
	public void controlEvent( ControlEvent e )
	{

		if( checkboxes.getItem( 0 ).value() == 0.0 )noSmooth();
		else smooth();
		
		
		//scale = sliderScale.value();

		tolerance = sliderTolerance.value();
		distance = sliderDistance.value();
		
		if( e.id() == 0 )drawBlobsAndEdges();
		
		Patch.friction = sliderFriction.value();
		Patch.damping = sliderDamping.value();
		
	}


	void drawBlobsAndEdges()
	{

		img = loadImage( "data/broken.png" );
		texture = loadImage( "data/broken_texture.jpg" );

		
//		size( (int)( img.width * scale ), (int)( img.height * scale ) );
		//System.out.println( (int)( img.width * scale ) +" "+ (int)( img.height * scale ) );
		
		Patch.texture = texture;
		Patch.scale = scale;
		
		
		patches = new ArrayList<Patch>();

		theBlobDetection = new BlobDetection(img.width, img.height);
		theBlobDetection.setPosDiscrimination(false);
		theBlobDetection.setThreshold(0.5f);
		theBlobDetection.computeBlobs(img.pixels);
		
		Blob b;
		EdgeVertex vertex;
		
		
		for (int n=0 ; n<theBlobDetection.getBlobNb() ; n++)
		{
			b = theBlobDetection.getBlob(n);
			
			if ( b != null )
			{
				//grab the blobs and outlines
				PVector[] vertices;
				ArrayList points = new ArrayList();
				for ( int m = 0; m < b.getEdgeNb(); m++ )
				{
					vertex = b.getEdgeVertexA( m );
					points.add( new PVector( vertex.x * width, vertex.y * height ) );
				}
				
				//simplifying the shapes
				vertices = Simplify.simplifyLang( points, (int)distance, tolerance );
				
				ArrayList Tpoints = new ArrayList();
				for( int i = 0; i < vertices.length; i++ )
				{
					Tpoints.add( new PVector( vertices[i].x, vertices[i].y ) );
				}

				//adding dots to the point set
				Rectangle rect = PolygonUtils.boundingBox( vertices );
				for ( int i =0; i < vertices.length * 3; i++ )
				{
					PVector p = new PVector( 	rect.x + random( rect.width ) , 
												rect.y + random( rect.height )); 
					if( PolygonUtils.contains( p, vertices ) )
					{
						Tpoints.add( p );
					}
				}
				
				//creating patches of triangles
				Patch patch = new Patch( vertices );
				ArrayList triangles = Triangulate.triangulate(Tpoints);
				for (int i = 0; i < triangles.size(); i++)
				{
					
				    Triangle t = (Triangle)triangles.get(i);
				   
				    biga.shapes2D.Triangle tri = new biga.shapes2D.Triangle( t.p1, t.p2, t.p3 );
				    
				    PVector p = biga.utils.TriangleUtils.centroid( tri );
	    
				    if( img.get( (int) ( p.x / scale ) , (int)( p.y / scale ) ) == color(255,255,255)
				    && 	PolygonUtils.contains( p, vertices ) )
				    {
				    	patch.addTriangle( tri );
				    }
				    
				}
				patches.add(patch);
				patch.reset();
			}
		}

		for ( Patch p: patches )
		{
		}
		
		
	}
	
	
}
