import java.util.ArrayList;

import processing.core.PApplet;
import processing.core.PImage;
import processing.core.PVector;
import biga.shapes2D.Circle;
import biga.shapes2D.Edge;
import biga.shapes2D.Triangle;
import biga.utils.GeomUtils;
import biga.utils.PolygonUtils;
import biga.utils.TriangleUtils;

public class Patch
{
	public static final int MOVE_TARGET = 0;
	public static final int MOVE_ELASTIC = 1;
	
	public static final int MOTION_NONE = 0;
	public static final int MOTION_RANDOM = 1;
	public static final int MOTION_FOLLOW = 2;
	public static final int MOTION_REPEL = 3;
	public static final int MOTION_ATTRACT = 4;
	
	
	public ArrayList<Triangle> triangles;
	public ArrayList<Triangle> targets;
	public ArrayList<Triangle> origins;
	public ArrayList<Triangle> velocities;
	private PVector[] outline;
//	private PVector[] outlinePoints;
	
	public static PImage texture;
	public static float scale;
	public static float friction = .2f;
	public static float damping = .8f;
	
	public Patch( PVector[] outline)
	{
		this.outline = outline;
//		outlinePoints = new PVector[ outline.length ];
		
		triangles = new ArrayList<Triangle>();
		targets = new ArrayList<Triangle>();
		origins = new ArrayList<Triangle>();
		velocities = new ArrayList<Triangle>();
		
	}
	
	public void addTriangle( Triangle tri )
	{
		triangles.add( tri );
		
		for ( int i = 0; i < outline.length; i++ )
		{
			for ( int j = 0; j < tri.points.length; j++ )
			{
				if( tri.points[ j ].x == outline[ i ].x
					&& tri.points[ j ].y == outline[ i ].y	)
				{
					outline[ i ] = tri.points[ j ];
				}
			}
		}
		
		
		if( texture != null )
		{
			PVector p = TriangleUtils.centroid(tri);
			tri.color = (int)texture.get( (int)( p.x / scale ), (int)( p.y / scale ) );
		}
		else
		{
			PVector _p0, _p1;
			int alp = 0;
			for (int i = 0; i < 3; i++ )
			{
				_p0 = tri.points[ i ];
				_p1 = tri.points[ (int) ( i + 1 ) % tri.points.length ];
				alp += (int) GeomUtils.distance( _p0, _p1 );
			}
			tri.color = (int)( Math.min( 0xFF, alp ) );
		}
		
		
		targets.add( tri.clone() );
		origins.add( tri.clone() );
		velocities.add( new Triangle() );
		
	}
	

	public void target()
	{
		Triangle tri, target, velocity;
		for( int i = 0; i < triangles.size(); i++ )
		{
			tri = (Triangle ) triangles.get( i );
			target = ( Triangle ) targets.get( i );
			for( int j = 0; j < tri.points.length; j++)
			{
		        tri.points[ j ].x += ( ( target.points[ j ].x - tri.points[ j ].x) * friction );
		        tri.points[ j ].y += ( ( target.points[ j ].y - tri.points[ j ].y) * friction );
			}
		}
	}
	
	public void elastic()
	{
		Triangle tri, target, velocity;
		for( int i = 0; i < triangles.size(); i++ )
		{
			tri = (Triangle ) triangles.get( i );
			target = ( Triangle ) targets.get( i );
			velocity = ( Triangle ) velocities.get( i );

			for( int j = 0; j < tri.points.length; j++ )
			{
		        velocity.points[ j ].x += ( (target.points[ j ].x-tri.points[ j ].x) * friction );
		        velocity.points[ j ].y += ( (target.points[ j ].y-tri.points[ j ].y) * friction );
		        
		        velocity.points[ j ].x *= damping;
		        velocity.points[ j ].y *= damping;
		        
		        tri.points[ j ].x += velocity.points[ j ].x;
		        tri.points[ j ].y += velocity.points[ j ].y;
			}
		}
	}
	
	public void reset()
	{
		for ( int i = 0; i < triangles.size(); i++ )
		{
			Triangle t = (Triangle ) targets.get( i );
			Triangle o = (Triangle ) origins.get( i );
			
			for( int j = 0; j < t.points.length; j++)
			{
		        t.points[ j ].x = o.points[ j ].x;
		        t.points[ j ].y = o.points[ j ].y;
			}
		}
	}

	public void targetPatch( Patch target )
	{
		
		int length = Math.min( triangles.size(), target.triangles.size() );
		Triangle t, o; 
		
		for ( int i = 0; i < triangles.size(); i++ )
		{
			
			int index = ( i % length ); 
			t = (Triangle ) targets.get( i );
			o = (Triangle ) target.origins.get( index );
			
			for( int j = 0; j < t.points.length; j++)
			{
		        t.points[ j ].x = o.points[ j ].x;
		        t.points[ j ].y = o.points[ j ].y;
			}
				
		}
	}
	
	public void attract( PVector point, double distance )
	{

		for( int i = 0; i < targets.size(); i++ )
		{
			Triangle t = ( Triangle ) targets.get( i );
			PVector p = TriangleUtils.centroid(t);
			double d = GeomUtils.distance( point, p );
			
			if( d < distance)
			{
				PVector offset = new PVector();
				double a = GeomUtils.angle( point, p ); 
				offset.x = (float) ( Math.cos( a ) * ( d / distance ) );
				offset.y = (float) ( Math.sin( a ) * ( d / distance ) );
				for( int j = 0; j < t.points.length; j++)
				{
			        t.points[ j ].x -= offset.x;
			        t.points[ j ].y -= offset.y;
				}
			}
			else
			{
				Triangle o = ( Triangle ) origins.get( i );
				for( int j = 0; j < t.points.length; j++)
				{
			        t.points[ j ].x += ( float ) ( o.points[ j ].x - t.points[ j ].x ) * friction;
			        t.points[ j ].y += ( float ) ( o.points[ j ].y - t.points[ j ].y ) * friction;
				}
			}
		}
	}
	public void repel( PVector point, double distance )
	{
		
		for( int i = 0; i < targets.size(); i++ )
		{
			Triangle t = ( Triangle ) targets.get( i );
			PVector p = TriangleUtils.centroid(t);
			double d = GeomUtils.distance( point, p );
			
			if( d < distance)
			{
				PVector offset = new PVector();
				double a = GeomUtils.angle( point, p ); 
				offset.x = (float) ( Math.cos( a ) * ( distance / d ) );
				offset.y = (float) ( Math.sin( a ) * ( distance / d ) );
				for( int j = 0; j < t.points.length; j++)
				{
					
			        t.points[ j ].x += offset.x;
			        t.points[ j ].y += offset.y;

				}
			}
			else
			{
				Triangle o = ( Triangle ) origins.get( i );
				for( int j = 0; j < t.points.length; j++)
				{
			        t.points[ j ].x += ( float ) ( o.points[ j ].x - t.points[ j ].x ) * friction;
			        t.points[ j ].y += ( float ) ( o.points[ j ].y - t.points[ j ].y ) * friction;
				}
			}
		}
	}

	public void shrink( PVector point, double distance )
	{

		for( int i = 0; i < targets.size(); i++ )
		{
			Triangle t = ( Triangle ) targets.get( i );
			PVector p = TriangleUtils.centroid(( Triangle ) origins.get( i ));
			double d = GeomUtils.distance( point, p );
			
			if( d < distance )
			{
				double a = GeomUtils.angle( point, p );
				double r = distance / d;
				for( int j = 0; j < t.points.length; j++)
				{
					a = GeomUtils.angle( t.points[ j ], p );
			        t.points[ j ].x = p.x;//(float) ( p.x + Math.cos( a ) * r );
			        t.points[ j ].y = p.y;//(float) ( p.y + Math.sin( a ) * r );
				}
			}
			else
			{
				Triangle o = ( Triangle ) origins.get( i );
				for( int j = 0; j < t.points.length; j++)
				{
			        t.points[ j ].x += ( float ) ( o.points[ j ].x - t.points[ j ].x ) * friction;
			        t.points[ j ].y += ( float ) ( o.points[ j ].y - t.points[ j ].y ) * friction;
				}
			}
		}
	}
	public void randomize( PVector destination, float amount )
	{
		for( Triangle t : targets )
		{
			for( int j = 0; j < t.points.length; j++)
			{
				t.points[ j ].x = destination.x + (float) ( ( Math.random() - .5 ) * amount );
		        t.points[ j ].y = destination.y + (float) ( ( Math.random() - .5 ) * amount );
			}
		}
	}

	public void randomMove( float amount )
	{
		for ( Triangle tri : triangles )
		{
			for( int j = 0; j < tri.points.length; j++)
			{
		        tri.points[ j ].x += (float) ( Math.random() - .5 ) * amount;
		        tri.points[ j ].y += (float) ( Math.random() - .5 ) * amount;
			}
		}
	}

	
	
	public void render( PApplet context )
	{
		for( Triangle t : triangles )
		{
			t.drawDistance( context, t.color, 0f );
		}
	}
	public void renderWireFrame( PApplet context )
	{
		for( Triangle t : triangles )
		{
			context.stroke(t.color>>16 & 0xFF, t.color>>8 & 0xFF, t.color & 0xFF);
			t.draw( context );
		}
	}
	public void renderIncircles( PApplet context, Boolean fill )
	{
		for( Triangle t : triangles )
		{
			Circle c = TriangleUtils.incircle(t);
			if( fill )
			{
				context.noStroke();
				context.fill(t.color>>16 & 0xFF, t.color>>8 & 0xFF, t.color & 0xFF);
			}
			else
			{
				context.noFill();
				context.stroke(t.color>>16 & 0xFF, t.color>>8 & 0xFF, t.color & 0xFF);
			}
			c.draw(context);
		}
	}
	
	public void renderVoronoi( PApplet context, Boolean fill )
	{
		context.strokeWeight(1);
		for( Triangle t : triangles )
		{
			context.noFill();
			context.stroke(t.color>>16 & 0xFF, t.color>>8 & 0xFF, t.color & 0xFF);
		
			Circle c = TriangleUtils.circumcircle(t);
			PVector ip = new PVector( c.x, c.y );
			if( PolygonUtils.contains(ip, outline ) )
			{
				context.line( ip.x, ip.y, t.edges[0].center().x, t.edges[0].center().y );
				context.line( ip.x, ip.y, t.edges[1].center().x, t.edges[1].center().y );
				context.line( ip.x, ip.y, t.edges[2].center().x, t.edges[2].center().y );
			}
		}
		
	}
	public void renderOutline( PApplet context, int color, int strokeWeight )
	{
		
		context.noFill();
		context.strokeWeight( strokeWeight );
		context.stroke( color>>16 & 0xFF, color>>8 & 0xFF, color & 0xFF);
		if( outline != null )
		{
			for ( int i =0; i < outline.length; i++ )
			{
				context.line( 	outline[ i ].x, 
								outline[ i ].y, 
								outline[ (int)( ( i + 1 ) % outline.length ) ].x, 
								outline[ (int)( ( i + 1 ) % outline.length ) ].y 	);
				
			}
		}
		context.strokeWeight( 1 );
	}
	
	public void resetElasticTargets()
	{
		for ( Triangle t : velocities )t = new Triangle();
	}
	
}
